@@ -1,5 +1,6 @@
 #include <dirent.h>
 #include <getopt.h>
+#include <signal.h>
 #include <stdbool.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -10,6 +11,14 @@
 
 #include "sb/counters.h"
 
+static const char *restore_screen_str = "\033[?1049l";
+
+static void exit_restore_screen(int n)
+{
+	write(STDOUT_FILENO, restore_screen_str, strlen(restore_screen_str));
+	exit(EXIT_SUCCESS);
+}
+
 static const u8 counters_to_stable_map[] = {
 #define x(n, id, ...)	[BCH_COUNTER_##n] = BCH_COUNTER_STABLE_##n,
 	BCH_PERSISTENT_COUNTERS()
@@ -57,6 +66,11 @@ static void fs_top(const char *path, bool human_readable)
 
 	unsigned interval_secs = 1;
 
+	struct sigaction act = { .sa_handler = exit_restore_screen };
+	sigaction(SIGINT, &act, NULL);
+
+	fputs("\033[?1049h", stdout);
+
 	while (true) {
 		sleep(interval_secs);
 		kfree(prev);
@@ -115,11 +129,13 @@ static void fs_top(const char *path, bool human_readable)
 	}
 
 	bcache_fs_close(fs);
+
+	fputs(restore_screen_str, stdout);
 }
 
 static void fs_top_usage(void)
 {
-	puts("bcachefs fs top - display runtime perfomance info\n"
+	puts("bcachefs fs top - display runtime performance info\n"
 	     "Usage: bcachefs fs top [OPTION]... <mountpoint>\n"
 	     "\n"
 	     "Options:\n"
@@ -127,6 +143,7 @@ static void fs_top_usage(void)
 	     "  -H, --help                   Display this help and exit\n"
 	     "\n"
 	     "Report bugs to <linux-bcachefs@vger.kernel.org>");
+	exit(EXIT_SUCCESS);
 }
 
 int cmd_fs_top(int argc, char *argv[])
