@@ -24,6 +24,8 @@ int cmd_fs_top(int argc, char *argv[]);
 
 int device_cmds(int argc, char *argv[]);
 
+int cmd_scrub(int argc, char *argv[]);
+int reconcile_cmds(int argc, char *argv[]);
 int data_cmds(int argc, char *argv[]);
 
 int cmd_unlock(int argc, char *argv[]);
@@ -34,6 +36,8 @@ int cmd_fsck(int argc, char *argv[]);
 int cmd_recovery_pass(int argc, char *argv[]);
 
 int cmd_dump(int argc, char *argv[]);
+int cmd_undump(int argc, char *argv[]);
+
 int cmd_list_journal(int argc, char *argv[]);
 int cmd_kill_btree_node(int argc, char *argv[]);
 
