@@ -441,6 +441,7 @@ int bch2_bucket_ref_update(struct btree_trans *trans, struct bch_dev *ca,
 				trans, stale_ptr_with_no_stale_ptrs_feature,
 				"stale cached ptr, but have no_stale_ptrs feature\n%s",
 				(bch2_bkey_val_to_text(&buf, c, k), buf.buf))) {
+			guard(memalloc_flags)(PF_MEMALLOC_NOFS);
 			guard(mutex)(&c->sb_lock);
 			c->disk_sb.sb->compat[0] &= ~cpu_to_le64(BIT_ULL(BCH_COMPAT_no_stale_ptrs));
 			bch2_write_super(c);
@@ -626,8 +627,7 @@ static int bch2_trigger_stripe_ptr(struct btree_trans *trans,
 	if (flags & BTREE_TRIGGER_transactional) {
 		struct bkey_i_stripe *s = bch2_bkey_get_mut_typed(trans,
 							BTREE_ID_stripes, POS(0, p.ec.idx),
-							BTREE_ITER_with_updates,
-							stripe);
+							0, stripe);
 		int ret = PTR_ERR_OR_ZERO(s);
 		if (unlikely(ret)) {
 			bch2_trans_inconsistent_on(bch2_err_matches(ret, ENOENT), trans,
@@ -961,7 +961,7 @@ int bch2_trans_mark_metadata_bucket(struct btree_trans *trans,
 	if (flags & BTREE_TRIGGER_gc)
 		return bch2_mark_metadata_bucket(trans, ca, b, type, sectors, flags);
 	else if (flags & BTREE_TRIGGER_transactional)
-		return commit_do(trans, NULL, NULL, 0,
+		return commit_do(trans, NULL, NULL, BCH_WATERMARK_btree,
 				 __bch2_trans_mark_metadata_bucket(trans, ca, b, type, sectors));
 	else
 		BUG();
