@@ -14,6 +14,7 @@
 #include "journal/journal.h"
 #include "journal/reclaim.h"
 
+#include "sb/counters.h"
 #include "sb/members.h"
 
 #include <linux/kthread.h>
@@ -72,7 +73,12 @@ void bch2_journal_set_watermark(struct journal *j)
 	if (track_event_change(&c->times[BCH_TIME_blocked_journal_low_on_space], low_on_space) ||
 	    track_event_change(&c->times[BCH_TIME_blocked_journal_low_on_pin], low_on_pin) ||
 	    track_event_change(&c->times[BCH_TIME_blocked_write_buffer_full], low_on_wb))
-		trace_and_count(c, journal_full, c);
+		event_inc_trace(c, journal_full, buf, ({
+			guard(printbuf_atomic)(&buf);
+			prt_printf(&buf, "low_on_space %u\n",	low_on_space);
+			prt_printf(&buf, "low_on_pin%u\n",	low_on_pin);
+			prt_printf(&buf, "low_on_wb%u\n",	low_on_wb);
+		}));
 
 	mod_bit(JOURNAL_low_on_space,	&j->flags, low_on_space);
 	mod_bit(JOURNAL_low_on_pin,	&j->flags, low_on_pin);
@@ -155,7 +161,7 @@ static struct journal_space __journal_space_available(struct journal *j, unsigne
 	size_t mem_limit = max_t(ssize_t, 0,
 			(totalram_pages() * PAGE_SIZE) / 4 - j->dirty_entry_bytes);
 
-	for_each_member_device_rcu(c, ca, &c->rw_devs[BCH_DATA_journal]) {
+	for_each_member_device_rcu(c, ca, &c->allocator.rw_devs[BCH_DATA_journal]) {
 		if (!ca->journal.nr)
 			continue;
 
@@ -203,7 +209,7 @@ void bch2_journal_space_available(struct journal *j)
 	lockdep_assert_held(&j->lock);
 	guard(rcu)();
 
-	for_each_member_device_rcu(c, ca, &c->rw_devs[BCH_DATA_journal]) {
+	for_each_member_device_rcu(c, ca, &c->allocator.rw_devs[BCH_DATA_journal]) {
 		struct journal_device *ja = &ca->journal;
 
 		if (!ja->nr)
@@ -225,14 +231,13 @@ void bch2_journal_space_available(struct journal *j)
 
 	j->can_discard = can_discard;
 
-	if (nr_online < metadata_replicas_required(c)) {
+	if (!nr_online) {
 		if (!(c->sb.features & BIT_ULL(BCH_FEATURE_small_image))) {
 			CLASS(printbuf, buf)();
 			guard(printbuf_atomic)(&buf);
-			prt_printf(&buf, "insufficient writeable journal devices available: have %u, need %u\n"
-				   "rw journal devs:", nr_online, metadata_replicas_required(c));
+			prt_printf(&buf, "no writeable journal devices available\n");
 
-			for_each_member_device_rcu(c, ca, &c->rw_devs[BCH_DATA_journal])
+			for_each_member_device_rcu(c, ca, &c->allocator.rw_devs[BCH_DATA_journal])
 				prt_printf(&buf, " %s", ca->name);
 
 			bch_err(c, "%s", buf.buf);
@@ -344,25 +349,38 @@ void bch2_journal_update_last_seq(struct journal *j)
 	}
 }
 
-void bch2_journal_update_last_seq_ondisk(struct journal *j, u64 last_seq_ondisk)
+int bch2_journal_update_last_seq_ondisk(struct journal *j, u64 last_seq_ondisk,
+					darray_replicas_entry_refs *refs)
 {
-	size_t dirty_entry_bytes = 0;
+	BUG_ON(last_seq_ondisk > j->pin.back);
 
-	scoped_guard(mutex, &j->last_seq_ondisk_lock)
-		while (j->last_seq_ondisk < last_seq_ondisk) {
-			struct journal_entry_pin_list *pin_list = journal_seq_pin(j, j->last_seq_ondisk);
+	for (u64 seq = j->last_seq_ondisk; seq < last_seq_ondisk; seq++) {
+		struct journal_entry_pin_list *pin_list = journal_seq_pin(j, seq);
 
-			dirty_entry_bytes += pin_list->bytes;
-			pin_list->bytes = 0;
+		if (pin_list->devs.e.nr_devs) {
+			replicas_entry_refs *e = darray_find_p(*refs, i,
+			    bch2_replicas_entry_eq(&i->replicas.e, &pin_list->devs.e));
 
-			j->last_seq_ondisk++;
+			if (e) {
+				e->nr_refs++;
+			} else {
+				try(darray_push_gfp(refs, ((replicas_entry_refs) {
+						    .nr_refs = 1,
+						    .replicas = pin_list->devs,
+				}), GFP_ATOMIC));
+			}
+
+			pin_list->devs.e.nr_devs = 0;
 		}
 
-	scoped_guard(spinlock, &j->lock) {
-		if (WARN_ON(j->dirty_entry_bytes < dirty_entry_bytes))
-			dirty_entry_bytes = j->dirty_entry_bytes;
-		j->dirty_entry_bytes -= dirty_entry_bytes;
+		if (WARN_ON(j->dirty_entry_bytes < pin_list->bytes))
+			pin_list->bytes = j->dirty_entry_bytes;
+
+		j->dirty_entry_bytes -= pin_list->bytes;
+		pin_list->bytes = 0;
 	}
+
+	return 0;
 }
 
 bool __bch2_journal_pin_put(struct journal *j, u64 seq)
@@ -495,6 +513,7 @@ void bch2_journal_pin_set(struct journal *j, u64 seq,
 
 	scoped_guard(spinlock, &j->lock) {
 		BUG_ON(seq < j->last_seq);
+		BUG_ON(seq > journal_cur_seq(j));
 
 		bool reclaim = __journal_pin_drop(j, pin);
 
@@ -690,11 +709,10 @@ static u64 journal_seq_to_flush(struct journal *j)
 static int __bch2_journal_reclaim(struct journal *j, bool direct, bool kicked)
 {
 	struct bch_fs *c = container_of(j, struct bch_fs, journal);
-	struct btree_cache *bc = &c->btree_cache;
+	struct bch_fs_btree_cache *bc = &c->btree.cache;
 	bool kthread = (current->flags & PF_KTHREAD) != 0;
 	u64 seq_to_flush;
 	size_t min_nr, min_key_cache, nr_flushed;
-	unsigned flags;
 	int ret = 0;
 
 	/*
@@ -704,7 +722,7 @@ static int __bch2_journal_reclaim(struct journal *j, bool direct, bool kicked)
 	 * we're holding the reclaim lock:
 	 */
 	lockdep_assert_held(&j->reclaim_lock);
-	flags = memalloc_noreclaim_save();
+	guard(memalloc_flags)(PF_MEMALLOC_NOFS);
 
 	do {
 		if (kthread && kthread_should_stop())
@@ -737,12 +755,15 @@ static int __bch2_journal_reclaim(struct journal *j, bool direct, bool kicked)
 
 		min_key_cache = min(bch2_nr_btree_keys_need_flush(c), (size_t) 128);
 
-		trace_and_count(c, journal_reclaim_start, c,
-				direct, kicked,
-				min_nr, min_key_cache,
-				atomic_long_read(&bc->nr_dirty), btree_cache_live,
-				atomic_long_read(&c->btree_key_cache.nr_dirty),
-				atomic_long_read(&c->btree_key_cache.nr_keys));
+		event_inc_trace(c, journal_reclaim_start, buf, ({
+			prt_printf(&buf, "direct %u kicked %u\n", direct, kicked);
+			prt_printf(&buf, "btree cache %lu/%zu min %zu\n",
+				   atomic_long_read(&bc->nr_dirty), btree_cache_live, min_nr);
+			prt_printf(&buf, "key cache %lu/%lu min %zu\n",
+				   atomic_long_read(&c->btree.key_cache.nr_dirty),
+				   atomic_long_read(&c->btree.key_cache.nr_keys),
+				   min_key_cache);
+		}));
 
 		nr_flushed = journal_flush_pins(j, seq_to_flush,
 						~0, 0,
@@ -752,14 +773,14 @@ static int __bch2_journal_reclaim(struct journal *j, bool direct, bool kicked)
 			j->nr_direct_reclaim += nr_flushed;
 		else
 			j->nr_background_reclaim += nr_flushed;
-		trace_and_count(c, journal_reclaim_finish, c, nr_flushed);
+
+		event_inc_trace(c, journal_reclaim_finish, buf,
+			prt_printf(&buf, "flushed %zu\n", nr_flushed));
 
 		if (nr_flushed)
 			wake_up(&j->reclaim_wait);
 	} while ((min_nr || min_key_cache) && nr_flushed && !direct);
 
-	memalloc_noreclaim_restore(flags);
-
 	return ret;
 }
 
@@ -973,39 +994,7 @@ int bch2_journal_flush_device_pins(struct journal *j, int dev_idx)
 
 	try(bch2_journal_error(j));
 
-	guard(mutex)(&c->replicas_gc_lock);
-	bch2_replicas_gc_start(c, 1 << BCH_DATA_journal);
-
-	/*
-	 * Now that we've populated replicas_gc, write to the journal to mark
-	 * active journal devices. This handles the case where the journal might
-	 * be empty. Otherwise we could clear all journal replicas and
-	 * temporarily put the fs into an unrecoverable state. Journal recovery
-	 * expects to find devices marked for journal data on unclean mount.
-	 */
-	int ret = bch2_journal_meta(&c->journal);
-	if (ret)
-		goto err;
-
-	seq = 0;
-	scoped_guard(spinlock, &j->lock)
-		while (!ret) {
-			seq = max(seq, j->last_seq);
-			if (seq > j->seq_ondisk)
-				break;
-
-			union bch_replicas_padded replicas;
-			memcpy(&replicas, &journal_seq_pin(j, seq)->devs, sizeof(replicas));
-			seq++;
-
-			if (replicas.e.nr_devs) {
-				spin_unlock(&j->lock);
-				ret = bch2_mark_replicas(c, &replicas.e);
-				spin_lock(&j->lock);
-			}
-		}
-err:
-	return bch2_replicas_gc_end(c, ret);
+	return 0;
 }
 
 bool bch2_journal_seq_pins_to_text(struct printbuf *out, struct journal *j, u64 *seq)
